from ai21.clients.studio.resources.maestro.maestro import AsyncMaestro
from ai21.clients.studio.resources.studio_conversational_rag import AsyncStudioConversationalRag
from ai21.clients.studio.resources.studio_resource import AsyncStudioResource
from ai21.http_client.async_http_client import AsyncAI21HTTPClient


class AsyncBeta(AsyncStudioResource):
    def __init__(self, client: AsyncAI21HTTPClient):
        super().__init__(client)

        self.conversational_rag = AsyncStudioConversationalRag(client)
        self.maestro = AsyncMaestro(client)



from __future__ import annotations

from typing import Dict, Any, Type

from pydantic import VERSION, BaseModel

IS_PYDANTIC_V2 = VERSION.startswith("2.")


def _to_dict(model_object: BaseModel, **kwargs) -> Dict[str, Any]:
    if IS_PYDANTIC_V2:
        return model_object.model_dump(**kwargs)

    return model_object.dict(**kwargs)


def _to_json(model_object: BaseModel, **kwargs) -> str:
    if IS_PYDANTIC_V2:
        return model_object.model_dump_json(**kwargs)

    return model_object.json(**kwargs)


def _from_dict(obj: "AI21BaseModel", obj_dict: Any, **kwargs) -> BaseModel:  # noqa: F821
    if IS_PYDANTIC_V2:
        return obj.model_validate(obj_dict, **kwargs)

    return obj.parse_obj(obj_dict, **kwargs)


def _from_json(obj: "AI21BaseModel", json_str: str, **kwargs) -> BaseModel:  # noqa: F821
    if IS_PYDANTIC_V2:
        return obj.model_validate_json(json_str, **kwargs)

    return obj.parse_raw(json_str, **kwargs)


def _to_schema(model_object: Type[BaseModel], **kwargs) -> Dict[str, Any]:
    if IS_PYDANTIC_V2:
        return model_object.model_json_schema(**kwargs)

    return model_object.schema(**kwargs)



from __future__ import annotations

import warnings

from abc import ABC
from typing import Any, Dict, List, Literal, Optional, Union

from ai21.models._pydantic_compatibility import _to_dict
from ai21.models.chat.chat_message import ChatMessageParam
from ai21.models.chat.document_schema import DocumentSchema
from ai21.models.chat.response_format import ResponseFormat
from ai21.models.chat.tool_defintions import ToolDefinition
from ai21.types import NotGiven
from ai21.utils.typing import remove_not_given


_MODEL_DEPRECATION_WARNING = """
The 'jamba-1.5-mini' and 'jamba-1.5-large' models are deprecated and will
be removed in a future version.
Please use jamba-mini-1.6-2025-03 or jamba-large-1.6-2025-03 instead.
"""


class BaseChatCompletions(ABC):
    _module_name = "chat/completions"

    def _check_model(self, model: Optional[str]) -> str:
        if not model:
            raise ValueError("model should be provided 'create' method call")

        if model in ["jamba-1.5-mini", "jamba-1.5-large"]:
            warnings.warn(
                _MODEL_DEPRECATION_WARNING,
                DeprecationWarning,
                stacklevel=3,
            )

        return model

    def _create_body(
        self,
        model: str,
        messages: List[ChatMessageParam],
        max_tokens: Optional[int] | NotGiven,
        temperature: Optional[float] | NotGiven,
        top_p: Optional[float] | NotGiven,
        stop: Optional[Union[str, List[str]]] | NotGiven,
        n: Optional[int] | NotGiven,
        stream: Literal[False] | Literal[True] | NotGiven,
        tools: List[ToolDefinition] | NotGiven,
        response_format: ResponseFormat | NotGiven,
        documents: List[DocumentSchema] | NotGiven,
        **kwargs: Any,
    ) -> Dict[str, Any]:
        return remove_not_given(
            {
                "model": model,
                "messages": [_to_dict(message) for message in messages],
                "temperature": temperature,
                "max_tokens": max_tokens,
                "top_p": top_p,
                "stop": stop,
                "n": n,
                "stream": stream,
                "tools": tools,
                "response_format": response_format,
                "documents": documents,
                **kwargs,
            }
        )



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import List, Any, Dict, Optional, TypeVar, Union

from ai21.clients.studio.resources.chat import ChatCompletions, AsyncChatCompletions
from ai21.clients.studio.resources.constants import (
    CHAT_DEFAULT_NUM_RESULTS,
    CHAT_DEFAULT_TEMPERATURE,
    CHAT_DEFAULT_MAX_TOKENS,
    CHAT_DEFAULT_MIN_TOKENS,
    CHAT_DEFAULT_TOP_P,
    CHAT_DEFAULT_TOP_K_RETURN,
)
from ai21.models import Penalty, ChatResponse, ChatMessage
from ai21.models._pydantic_compatibility import _to_dict
from ai21.types import NotGiven, NOT_GIVEN
from ai21.utils.typing import remove_not_given

_ChatCompletionsT = TypeVar("_ChatCompletionsT", bound=Union[ChatCompletions, AsyncChatCompletions])


class Chat(ABC):
    _module_name = "chat"

    @abstractmethod
    def create(
        self,
        model: str,
        messages: List[ChatMessage],
        system: str,
        *,
        num_results: Optional[int] = CHAT_DEFAULT_NUM_RESULTS,
        temperature: Optional[float] = CHAT_DEFAULT_TEMPERATURE,
        max_tokens: Optional[int] = CHAT_DEFAULT_MAX_TOKENS,
        min_tokens: Optional[int] = CHAT_DEFAULT_MIN_TOKENS,
        top_p: Optional[float] = CHAT_DEFAULT_TOP_P,
        top_k_return: Optional[int] = CHAT_DEFAULT_TOP_K_RETURN,
        stop_sequences: Optional[List[str]] | NotGiven = NOT_GIVEN,
        frequency_penalty: Optional[Penalty] | NotGiven = NOT_GIVEN,
        presence_penalty: Optional[Penalty] | NotGiven = NOT_GIVEN,
        count_penalty: Optional[Penalty] | NotGiven = NOT_GIVEN,
        **kwargs,
    ) -> ChatResponse:
        """

        :param model: model type you wish to interact with
        :param messages: A sequence of messages ingested by the model, which then returns the assistant's response
        :param system: Offers the model overarching guidance on its response approach, encapsulating context, tone,
         guardrails, and more
        :param max_tokens: The maximum number of tokens to generate per result
        :param num_results: Number of completions to sample and return.
        :param min_tokens: The minimum number of tokens to generate per result.
        :param temperature: A value controlling the "creativity" of the model's responses.
        :param top_p: A value controlling the diversity of the model's responses.
        :param top_k_return: The number of top-scoring tokens to consider for each generation step.
        :param stop_sequences: Stops decoding if any of the strings is generated
        :param frequency_penalty: A penalty applied to tokens that are frequently generated.
        :param presence_penalty:  A penalty applied to tokens that are already present in the prompt.
        :param count_penalty: A penalty applied to tokens based on their frequency in the generated responses
        :param kwargs:
        :return:
        """
        pass

    @property
    @abstractmethod
    def completions(self) -> _ChatCompletionsT:
        pass

    def _create_body(
        self,
        model: str,
        messages: List[ChatMessage],
        system: str,
        num_results: Optional[int] = CHAT_DEFAULT_NUM_RESULTS,
        temperature: Optional[float] = CHAT_DEFAULT_TEMPERATURE,
        max_tokens: Optional[int] = CHAT_DEFAULT_MAX_TOKENS,
        min_tokens: Optional[int] = CHAT_DEFAULT_MIN_TOKENS,
        top_p: Optional[float] = CHAT_DEFAULT_TOP_P,
        top_k_return: Optional[int] = CHAT_DEFAULT_TOP_K_RETURN,
        stop_sequences: Optional[List[str]] | NotGiven = NOT_GIVEN,
        frequency_penalty: Optional[Penalty] | NotGiven = NOT_GIVEN,
        presence_penalty: Optional[Penalty] | NotGiven = NOT_GIVEN,
        count_penalty: Optional[Penalty] | NotGiven = NOT_GIVEN,
        **kwargs,
    ) -> Dict[str, Any]:
        return remove_not_given(
            {
                "model": model,
                "system": system,
                "messages": [_to_dict(message) for message in messages],
                "temperature": temperature,
                "maxTokens": max_tokens,
                "minTokens": min_tokens,
                "numResults": num_results,
                "topP": top_p,
                "topKReturn": top_k_return,
                "stopSequences": stop_sequences,
                "frequencyPenalty": (NOT_GIVEN if frequency_penalty is NOT_GIVEN else _to_dict(frequency_penalty)),
                "presencePenalty": (NOT_GIVEN if presence_penalty is NOT_GIVEN else _to_dict(presence_penalty)),
                "countPenalty": (NOT_GIVEN if count_penalty is NOT_GIVEN else _to_dict(count_penalty)),
                **kwargs,
            }
        )



from __future__ import annotations

from typing import Optional, Tuple

import google.auth
from google.auth.credentials import Credentials
from google.auth.transport.requests import Request
from google.auth.exceptions import DefaultCredentialsError

from ai21.errors import CredentialsError


class GCPAuthorization:
    def get_gcp_credentials(
        self,
        project_id: Optional[str] = None,
    ) -> Tuple[Credentials, str]:
        try:
            credentials, loaded_project_id = google.auth.default(
                scopes=["https://www.googleapis.com/auth/cloud-platform"],
            )
        except DefaultCredentialsError as e:
            raise CredentialsError(provider_name="GCP", error_message=str(e))

        if project_id is not None and project_id != loaded_project_id:
            raise ValueError("Mismatch between credentials project id and 'project_id'")

        project_id = project_id or loaded_project_id

        if project_id is None:
            raise ValueError("Could not get project_id for GCP project")

        if not isinstance(project_id, str):
            raise ValueError(f"Variable project_id must be a string, got {type(project_id)} instead")

        return credentials, project_id

    def _get_gcp_request(self) -> Request:
        return Request()

    def refresh_auth(self, credentials: Credentials) -> None:
        request = self._get_gcp_request()
        credentials.refresh(request)



from typing import Optional, List

from ai21.models.ai21_base_model import AI21BaseModel
from ai21.models.logprobs import Logprobs
from ai21.models.usage_info import UsageInfo
from .chat_message import AssistantMessage


class ChatCompletionResponseChoice(AI21BaseModel):
    index: int
    message: AssistantMessage
    logprobs: Optional[Logprobs] = None
    finish_reason: Optional[str] = None


class ChatCompletionResponse(AI21BaseModel):
    id: str
    choices: List[ChatCompletionResponseChoice]
    usage: UsageInfo



from __future__ import annotations

from .chat_completion_chunk import ChatCompletionChunk, ChoicesChunk, ChoiceDelta
from .chat_completion_response import ChatCompletionResponse
from .chat_completion_response import ChatCompletionResponseChoice
from .chat_message import ChatMessage, AssistantMessage, ToolMessage, UserMessage, SystemMessage, ChatMessageParam
from .document_schema import DocumentSchema
from .function_tool_definition import FunctionToolDefinition
from .response_format import ResponseFormat
from .role_type import RoleType as RoleType
from .tool_call import ToolCall
from .tool_defintions import ToolDefinition
from .tool_function import ToolFunction
from .tool_parameters import ToolParameters

__all__ = [
    "ChatCompletionResponse",
    "ChatCompletionResponseChoice",
    "ChatMessage",
    "RoleType",
    "ChatCompletionChunk",
    "ChoicesChunk",
    "ChoiceDelta",
    "AssistantMessage",
    "ToolMessage",
    "UserMessage",
    "SystemMessage",
    "ChatMessageParam",
    "DocumentSchema",
    "FunctionToolDefinition",
    "ResponseFormat",
    "ToolCall",
    "ToolDefinition",
    "ToolFunction",
    "ToolParameters",
]



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import List, Dict, Any

from ai21.models.chat import ChatMessage
from ai21.models.maestro.run import (
    Tool,
    ToolResources,
    RunResponse,
    DEFAULT_RUN_POLL_INTERVAL,
    DEFAULT_RUN_POLL_TIMEOUT,
    Requirement,
    Budget,
)
from ai21.types import NOT_GIVEN, NotGiven
from ai21.utils.typing import remove_not_given


class BaseMaestroRun(ABC):
    _module_name = "maestro/runs"

    def _create_body(
        self,
        *,
        input: str | List[ChatMessage],
        models: List[str] | NotGiven,
        tools: List[Tool] | NotGiven,
        tool_resources: ToolResources | NotGiven,
        context: Dict[str, Any] | NotGiven,
        requirements: List[Requirement] | NotGiven,
        budget: Budget | NotGiven,
        **kwargs,
    ) -> dict:
        return remove_not_given(
            {
                "input": input,
                "models": models,
                "tools": tools,
                "tool_resources": tool_resources,
                "context": context,
                "requirements": requirements,
                "budget": budget,
                **kwargs,
            }
        )

    @abstractmethod
    def create(
        self,
        *,
        input: str | List[ChatMessage],
        models: List[str] | NotGiven = NOT_GIVEN,
        tools: List[Tool] | NotGiven = NOT_GIVEN,
        tool_resources: ToolResources | NotGiven = NOT_GIVEN,
        context: Dict[str, Any] | NotGiven = NOT_GIVEN,
        requirements: List[Requirement] | NotGiven = NOT_GIVEN,
        budget: Budget | NotGiven = NOT_GIVEN,
        **kwargs,
    ) -> RunResponse:
        pass

    @abstractmethod
    def retrieve(self, run_id: str) -> RunResponse:
        pass

    @abstractmethod
    def _poll_for_status(self, *, run_id: str, poll_interval: float, poll_timeout: float) -> RunResponse:
        pass

    @abstractmethod
    def create_and_poll(
        self,
        *,
        input: str | List[ChatMessage],
        models: List[str] | NotGiven = NOT_GIVEN,
        tools: List[Tool] | NotGiven = NOT_GIVEN,
        tool_resources: ToolResources | NotGiven = NOT_GIVEN,
        context: Dict[str, Any] | NotGiven = NOT_GIVEN,
        requirements: List[Requirement] | NotGiven = NOT_GIVEN,
        budget: Budget | NotGiven = NOT_GIVEN,
        poll_interval_sec: float = DEFAULT_RUN_POLL_INTERVAL,
        poll_timeout_sec: float = DEFAULT_RUN_POLL_TIMEOUT,
        **kwargs,
    ) -> RunResponse:
        pass



from ai21.clients.common.maestro.maestro import BaseMaestro
from ai21.clients.studio.resources.maestro.run import MaestroRun, AsyncMaestroRun
from ai21.clients.studio.resources.studio_resource import StudioResource, AsyncStudioResource
from ai21.http_client.async_http_client import AsyncAI21HTTPClient
from ai21.http_client.http_client import AI21HTTPClient


class Maestro(StudioResource, BaseMaestro):
    def __init__(self, client: AI21HTTPClient):
        super().__init__(client)

        self.runs = MaestroRun(client)


class AsyncMaestro(AsyncStudioResource, BaseMaestro):
    def __init__(self, client: AsyncAI21HTTPClient):
        super().__init__(client)

        self.runs = AsyncMaestroRun(client)



from typing import Optional, Dict

from botocore.auth import SigV4Auth
from botocore.awsrequest import AWSRequest
import boto3


class AWSAuthorization:
    def __init__(self, aws_session: boto3.Session):
        self._aws_session = aws_session

    def get_auth_headers(
        self,
        *,
        url: str,
        service_name: str,
        method: str,
        data: Optional[str],
    ) -> Dict[str, str]:
        request = AWSRequest(method=method, url=url, data=data)
        credentials = self._aws_session.get_credentials()

        signer = SigV4Auth(
            credentials=credentials, service_name=service_name, region_name=self._aws_session.region_name
        )
        signer.add_auth(request)

        prepped = request.prepare()

        return {key: value for key, value in dict(prepped.headers).items() if value is not None}



from typing import Optional, List

from pydantic import Field

from ai21.models.ai21_base_model import AI21BaseModel
from ai21.models.chat.role_type import RoleType


class FinishReason(AI21BaseModel):
    reason: str
    length: Optional[int] = None
    sequence: Optional[str] = None


class ChatOutput(AI21BaseModel):
    text: str
    role: RoleType
    finish_reason: Optional[FinishReason] = Field(default=None, alias="finishReason")


class ChatResponse(AI21BaseModel):
    outputs: List[ChatOutput]



from __future__ import annotations

import asyncio
import time
from typing import Any, List, Dict

from ai21.clients.common.maestro.run import BaseMaestroRun
from ai21.clients.studio.resources.studio_resource import StudioResource, AsyncStudioResource
from ai21.models.chat import ChatMessage
from ai21.models.maestro.run import (
    Tool,
    ToolResources,
    RunResponse,
    TERMINATED_RUN_STATUSES,
    DEFAULT_RUN_POLL_INTERVAL,
    DEFAULT_RUN_POLL_TIMEOUT,
    Requirement,
    Budget,
)
from ai21.types import NotGiven, NOT_GIVEN


class MaestroRun(StudioResource, BaseMaestroRun):
    def create(
        self,
        *,
        input: str | List[ChatMessage],
        models: List[str] | NotGiven = NOT_GIVEN,
        tools: List[Tool] | NotGiven = NOT_GIVEN,
        tool_resources: ToolResources | NotGiven = NOT_GIVEN,
        context: Dict[str, Any] | NotGiven = NOT_GIVEN,
        requirements: List[Requirement] | NotGiven = NOT_GIVEN,
        budget: Budget | NotGiven = NOT_GIVEN,
        **kwargs,
    ) -> RunResponse:
        body = self._create_body(
            input=input,
            models=models,
            tools=tools,
            tool_resources=tool_resources,
            context=context,
            requirements=requirements,
            budget=budget,
            **kwargs,
        )

        return self._post(path=f"/{self._module_name}", body=body, response_cls=RunResponse)

    def retrieve(
        self,
        run_id: str,
    ) -> RunResponse:
        return self._get(path=f"/{self._module_name}/{run_id}", response_cls=RunResponse)

    def _poll_for_status(self, *, run_id: str, poll_interval: float, poll_timeout: float) -> RunResponse:
        start_time = time.time()

        while True:
            run = self.retrieve(run_id)

            if run.status in TERMINATED_RUN_STATUSES:
                return run

            if (time.time() - start_time) >= poll_timeout:
                return run

            time.sleep(poll_interval)

    def create_and_poll(
        self,
        *,
        input: str | List[ChatMessage],
        models: List[str] | NotGiven = NOT_GIVEN,
        tools: List[Tool] | NotGiven = NOT_GIVEN,
        tool_resources: ToolResources | NotGiven = NOT_GIVEN,
        context: Dict[str, Any] | NotGiven = NOT_GIVEN,
        requirements: List[Requirement] | NotGiven = NOT_GIVEN,
        budget: Budget | NotGiven = NOT_GIVEN,
        poll_interval_sec: float = DEFAULT_RUN_POLL_INTERVAL,
        poll_timeout_sec: float = DEFAULT_RUN_POLL_TIMEOUT,
        **kwargs,
    ) -> RunResponse:
        run = self.create(
            input=input,
            models=models,
            tools=tools,
            tool_resources=tool_resources,
            context=context,
            requirements=requirements,
            budget=budget,
            **kwargs,
        )

        return self._poll_for_status(run_id=run.id, poll_interval=poll_interval_sec, poll_timeout=poll_timeout_sec)


class AsyncMaestroRun(AsyncStudioResource, BaseMaestroRun):
    async def create(
        self,
        *,
        input: str | List[ChatMessage],
        models: List[str] | NotGiven = NOT_GIVEN,
        tools: List[Tool] | NotGiven = NOT_GIVEN,
        tool_resources: ToolResources | NotGiven = NOT_GIVEN,
        context: Dict[str, Any] | NotGiven = NOT_GIVEN,
        requirements: List[Requirement] | NotGiven = NOT_GIVEN,
        budget: Budget | NotGiven = NOT_GIVEN,
        **kwargs,
    ) -> RunResponse:
        body = self._create_body(
            input=input,
            models=models,
            tools=tools,
            tool_resources=tool_resources,
            context=context,
            requirements=requirements,
            budget=budget,
            **kwargs,
        )

        return await self._post(path=f"/{self._module_name}", body=body, response_cls=RunResponse)

    async def retrieve(
        self,
        run_id: str,
    ) -> RunResponse:
        return await self._get(path=f"/{self._module_name}/{run_id}", response_cls=RunResponse)

    async def _poll_for_status(self, *, run_id: str, poll_interval: float, poll_timeout: float) -> RunResponse:
        start_time = time.time()

        while True:
            run = await self.retrieve(run_id)

            if run.status in TERMINATED_RUN_STATUSES:
                return run

            if (time.time() - start_time) >= poll_timeout:
                return run

            await asyncio.sleep(poll_interval)

    async def create_and_poll(
        self,
        *,
        input: str | List[ChatMessage],
        models: List[str] | NotGiven = NOT_GIVEN,
        tools: List[Tool] | NotGiven = NOT_GIVEN,
        tool_resources: ToolResources | NotGiven = NOT_GIVEN,
        context: Dict[str, Any] | NotGiven = NOT_GIVEN,
        requirements: List[Requirement] | NotGiven = NOT_GIVEN,
        budget: Budget | NotGiven = NOT_GIVEN,
        poll_interval_sec: float = DEFAULT_RUN_POLL_INTERVAL,
        poll_timeout_sec: float = DEFAULT_RUN_POLL_TIMEOUT,
        **kwargs,
    ) -> RunResponse:
        run = await self.create(
            input=input,
            models=models,
            tools=tools,
            tool_resources=tool_resources,
            context=context,
            requirements=requirements,
            budget=budget,
            **kwargs,
        )

        return await self._poll_for_status(
            run_id=run.id, poll_interval=poll_interval_sec, poll_timeout=poll_timeout_sec
        )



from typing import TypedDict, Literal, List, Optional, Any, Set, Dict, Type, Union

from pydantic import BaseModel

from ai21.models.ai21_base_model import AI21BaseModel

Budget = Literal["low", "medium", "high"]
Role = Literal["user", "assistant"]
RunStatus = Literal["completed", "failed", "in_progress", "requires_action"]
ToolType = Literal["file_search", "web_search"]
PrimitiveTypes = Union[Type[str], Type[int], Type[float], Type[bool]]
PrimitiveLists = Type[List[PrimitiveTypes]]
OutputType = Union[Type[BaseModel], PrimitiveTypes, Dict[str, Any]]

DEFAULT_RUN_POLL_INTERVAL: float = 1  # seconds
DEFAULT_RUN_POLL_TIMEOUT: float = 120  # seconds
TERMINATED_RUN_STATUSES: Set[RunStatus] = {"completed", "failed", "requires_action"}


class Tool(TypedDict):
    type: ToolType


class FileSearchToolResource(TypedDict, total=False):
    retrieval_similarity_threshold: Optional[float]
    labels: Optional[List[str]]
    labels_filter_mode: Optional[Literal["AND", "OR"]]
    labels_filter: Optional[dict]
    file_ids: Optional[List[str]]
    retrieval_strategy: Optional[str]
    max_neighbors: Optional[int]


class WebSearchToolResource(TypedDict, total=False):
    urls: Optional[List[str]]


class ToolResources(TypedDict, total=False):
    file_search: Optional[FileSearchToolResource]
    web_search: Optional[WebSearchToolResource]


class Requirement(TypedDict):
    name: str
    description: str


class RunResponse(AI21BaseModel):
    id: str
    status: RunStatus
    result: Any



from __future__ import annotations

from typing import Optional, Dict, Any

import httpx
from google.auth.credentials import Credentials as GCPCredentials

from ai21.clients.studio.resources.studio_chat import StudioChat, AsyncStudioChat
from ai21.clients.vertex.gcp_authorization import GCPAuthorization
from ai21.http_client.async_http_client import AsyncAI21HTTPClient
from ai21.http_client.http_client import AI21HTTPClient
from ai21.models.request_options import RequestOptions

_DEFAULT_GCP_REGION = "us-central1"
_VERTEX_BASE_URL_FORMAT = "https://{region}-aiplatform.googleapis.com/v1"
_VERTEX_PATH_FORMAT = "/projects/{project_id}/locations/{region}/publishers/ai21/models/{model}:{endpoint}"


class BaseAI21VertexClient:
    def __init__(
        self,
        region: Optional[str] = None,
        project_id: Optional[str] = None,
        access_token: Optional[str] = None,
        credentials: Optional[GCPCredentials] = None,
    ):
        if access_token is not None and project_id is None:
            raise ValueError("Field project_id is required when setting access_token")
        self._region = region or _DEFAULT_GCP_REGION
        self._access_token = access_token
        self._project_id = project_id
        self._credentials = credentials
        self._gcp_auth = GCPAuthorization()

    def _get_base_url(self) -> str:
        return _VERTEX_BASE_URL_FORMAT.format(region=self._region)

    def _get_access_token(self) -> str:
        if self._access_token is not None:
            return self._access_token

        if self._credentials is None:
            self._credentials, self._project_id = self._gcp_auth.get_gcp_credentials(
                project_id=self._project_id,
            )

        if self._credentials is None:
            raise ValueError("Could not get credentials for GCP project")

        self._gcp_auth.refresh_auth(self._credentials)

        if self._credentials.token is None:
            raise RuntimeError(f"Could not get access token for GCP project {self._project_id}")

        return self._credentials.token

    def _build_path(
        self,
        project_id: str,
        region: str,
        model: str,
        endpoint: str,
    ) -> str:
        return _VERTEX_PATH_FORMAT.format(
            project_id=project_id,
            region=region,
            model=model,
            endpoint=endpoint,
        )

    def _get_authorization_header(self) -> Dict[str, Any]:
        access_token = self._get_access_token()
        return {"Authorization": f"Bearer {access_token}"}


class AI21VertexClient(BaseAI21VertexClient, AI21HTTPClient):
    def __init__(
        self,
        region: Optional[str] = None,
        project_id: Optional[str] = None,
        base_url: Optional[str] = None,
        access_token: Optional[str] = None,
        credentials: Optional[GCPCredentials] = None,
        headers: Dict[str, str] | None = None,
        timeout_sec: Optional[float] = None,
        num_retries: Optional[int] = None,
        http_client: Optional[httpx.Client] = None,
    ):
        BaseAI21VertexClient.__init__(
            self,
            region=region,
            project_id=project_id,
            access_token=access_token,
            credentials=credentials,
        )

        if base_url is None:
            base_url = self._get_base_url()

        AI21HTTPClient.__init__(
            self,
            base_url=base_url,
            timeout_sec=timeout_sec,
            num_retries=num_retries,
            headers=headers,
            client=http_client,
            requires_api_key=False,
        )

        self.chat = StudioChat(self)
        # Override the chat.create method to match the completions endpoint,
        # so it wouldn't get to the old J2 completion endpoint
        self.chat.create = self.chat.completions.create

    def _build_request(self, options: RequestOptions) -> httpx.Request:
        options = self._prepare_options(options)

        return super()._build_request(options)

    def _prepare_options(self, options: RequestOptions) -> RequestOptions:
        body = options.body

        model = body.pop("model")
        stream = body.pop("stream", False)
        endpoint = "streamRawPredict" if stream else "rawPredict"
        headers = self._prepare_headers()
        path = self._build_path(
            project_id=self._project_id,
            region=self._region,
            model=model,
            endpoint=endpoint,
        )

        return options.replace(
            body=body,
            path=path,
            headers=headers,
        )

    def _prepare_headers(self) -> Dict[str, Any]:
        return self._get_authorization_header()


class AsyncAI21VertexClient(BaseAI21VertexClient, AsyncAI21HTTPClient):
    def __init__(
        self,
        region: Optional[str] = None,
        project_id: Optional[str] = None,
        base_url: Optional[str] = None,
        access_token: Optional[str] = None,
        credentials: Optional[GCPCredentials] = None,
        headers: Dict[str, str] | None = None,
        timeout_sec: Optional[float] = None,
        num_retries: Optional[int] = None,
        http_client: Optional[httpx.AsyncClient] = None,
    ):
        BaseAI21VertexClient.__init__(
            self,
            region=region,
            project_id=project_id,
            access_token=access_token,
            credentials=credentials,
        )

        if base_url is None:
            base_url = self._get_base_url()

        AsyncAI21HTTPClient.__init__(
            self,
            base_url=base_url,
            timeout_sec=timeout_sec,
            num_retries=num_retries,
            headers=headers,
            client=http_client,
            requires_api_key=False,
        )

        self.chat = AsyncStudioChat(self)
        # Override the chat.create method to match the completions endpoint,
        # so it wouldn't get to the old J2 completion endpoint
        self.chat.create = self.chat.completions.create

    def _build_request(self, options: RequestOptions) -> httpx.Request:
        options = self._prepare_options(options)

        return super()._build_request(options)

    def _prepare_options(self, options: RequestOptions) -> RequestOptions:
        body = options.body

        model = body.pop("model")
        stream = body.pop("stream", False)
        endpoint = "streamRawPredict" if stream else "rawPredict"
        headers = self._prepare_headers()
        path = self._build_path(
            project_id=self._project_id,
            region=self._region,
            model=model,
            endpoint=endpoint,
        )

        return options.replace(
            body=body,
            path=path,
            headers=headers,
        )

    def _prepare_headers(self) -> Dict[str, Any]:
        return self._get_authorization_header()



import logging
from typing import Any, Optional, Dict, BinaryIO
from tenacity import retry, retry_if_result, stop_after_attempt, wait_exponential, RetryError
from httpx import ConnectError

import httpx

from ai21.logger import logger
from ai21.models.request_options import RequestOptions
from ai21.stream.async_stream import AsyncStream
from ai21.http_client.base_http_client import (
    BaseHttpClient,
    handle_non_success_response,
    RETRY_BACK_OFF_FACTOR,
    TIME_BETWEEN_RETRIES,
)
from ai21.stream.stream_commons import _SSEDecoder

_logger = logging.getLogger(__name__)


def _requests_retry_async_session(retries: int) -> httpx.AsyncHTTPTransport:
    return httpx.AsyncHTTPTransport(
        retries=retries,
    )


class AsyncAI21HTTPClient(BaseHttpClient[httpx.AsyncClient, AsyncStream[Any]]):
    def __init__(
        self,
        api_key: Optional[str] = None,
        requires_api_key: bool = True,
        client: Optional[httpx.AsyncClient] = None,
        timeout_sec: int = None,
        num_retries: int = None,
        headers: Dict = None,
        via: Optional[str] = None,
        base_url: Optional[str] = None,
    ):
        super().__init__(
            api_key=api_key,
            base_url=base_url,
            requires_api_key=requires_api_key,
            timeout_sec=timeout_sec,
            num_retries=num_retries,
            headers=headers,
            via=via,
        )
        self._client = self._init_client(client)
        self._headers = self._build_headers(passed_headers=headers)

        # Since we can't use the retry decorator on a method of a class as we can't access class attributes,
        # we have to wrap the method in a function
        self._request = retry(
            wait=wait_exponential(multiplier=RETRY_BACK_OFF_FACTOR, min=TIME_BETWEEN_RETRIES),
            retry=retry_if_result(self._should_retry),
            stop=stop_after_attempt(self._num_retries),
        )(self._run_request)
        self._streaming_decoder = _SSEDecoder()

    async def execute_http_request(
        self,
        method: str,
        path: Optional[str] = None,
        params: Optional[Dict] = None,
        body: Optional[Dict] = None,
        stream: bool = False,
        files: Optional[Dict[str, BinaryIO]] = None,
        extra_headers: Optional[Dict] = None,
    ) -> httpx.Response:
        headers = {**self._headers, **extra_headers} if extra_headers is not None else self._headers

        options = RequestOptions(
            path=path,
            body=body,
            method=method,
            stream=stream,
            params=params,
            files=files,
            headers=headers,
            timeout=self._timeout_sec,
            url=self._base_url,
        )

        try:
            response = await self._request(options=options)
        except RetryError as retry_error:
            last_attempt = retry_error.last_attempt

            if last_attempt.failed:
                raise last_attempt.exception()
            else:
                response = last_attempt.result()

        except ConnectError as connection_error:
            logger.error(f"Calling {method} {self._base_url} failed with ConnectionError: {connection_error}")
            raise connection_error
        except Exception as exception:
            logger.error(f"Calling {method} {self._base_url} failed with Exception: {exception}")
            raise exception

        if response.status_code != httpx.codes.OK:
            logger.error(
                f"Calling {method} {self._base_url} failed with a non-200 response code: {response.status_code}"
            )

            if stream:
                details = self._extract_streaming_error_details(response)
                handle_non_success_response(response.status_code, details)
            else:
                handle_non_success_response(response.status_code, response.text)

        return response

    async def _run_request(self, options: RequestOptions) -> httpx.Response:
        request = self._build_request(options)

        _logger.debug(f"Calling {request.method} {request.url} {request.headers}, {options.body}")
        return await self._client.send(request=request, stream=options.stream)

    def _init_client(self, client: Optional[httpx.AsyncClient]) -> httpx.AsyncClient:
        if client is not None:
            return client

        if self._apply_retry_policy:
            return httpx.AsyncClient(transport=_requests_retry_async_session(retries=self._num_retries))

        return httpx.AsyncClient()

    def _get_streaming_decoder(self) -> _SSEDecoder:
        return self._streaming_decoder



from typing import Any, Union, List

import httpx
from typing_extensions import Literal, TypeVar, TYPE_CHECKING

from ai21.stream.async_stream import AsyncStream
from ai21.stream.stream import Stream

if TYPE_CHECKING:
    from ai21.models.ai21_base_model import AI21BaseModel  # noqa

ResponseT = TypeVar("_ResponseT", bound=Union["AI21BaseModel", str, httpx.Response, List[Any]])
StreamT = TypeVar("_StreamT", bound=Stream[Any])
AsyncStreamT = TypeVar("_AsyncStreamT", bound=AsyncStream[Any])


# Sentinel class used until PEP 0661 is accepted
class NotGiven:
    """
    A sentinel singleton class used to distinguish omitted keyword arguments
    from those passed in with the value None (which may have different behavior).

    For example:

    ```py
    def get(timeout: Union[int, NotGiven, None] = NotGiven()) -> Response:
        ...


    get(timeout=1)  # 1s timeout
    get(timeout=None)  # No timeout
    get()  # Default timeout behavior, which may not be statically known at the method definition.
    ```
    """

    def __bool__(self) -> Literal[False]:
        return False

    def __repr__(self) -> str:
        return "NOT_GIVEN"


NOT_GIVEN = NotGiven()



import json

from typing import Any, Dict, Optional

import boto3
import httpx

from ai21 import AI21APIError
from ai21.ai21_env_config import AI21EnvConfig
from ai21.clients.aws.aws_authorization import AWSAuthorization
from ai21.clients.bedrock._stream_decoder import _AWSEventStreamDecoder
from ai21.clients.studio.resources.studio_chat import AsyncStudioChat, StudioChat
from ai21.errors import AccessDenied, APITimeoutError, ModelErrorException, NotFound
from ai21.http_client.async_http_client import AsyncAI21HTTPClient
from ai21.http_client.http_client import AI21HTTPClient
from ai21.models.request_options import RequestOptions


_BEDROCK_URL_FORMAT = "https://bedrock-runtime.{region}.amazonaws.com"


def _handle_bedrock_error(aws_error: AI21APIError) -> None:
    status_code = aws_error.status_code
    if status_code == 403:
        raise AccessDenied(details=aws_error.details)

    if status_code == 404:
        raise NotFound(details=aws_error.details)

    if status_code == 408:
        raise APITimeoutError(details=aws_error.details)

    if status_code == 424:
        raise ModelErrorException(details=aws_error.details)

    raise aws_error


class BaseBedrockClient:
    def __init__(self, region: str, session: Optional[boto3.Session]):
        self._aws_auth = AWSAuthorization(aws_session=session or boto3.Session(region_name=region))
        self._streaming_decoder = _AWSEventStreamDecoder()

    def _prepare_options(self, options: RequestOptions) -> RequestOptions:
        body = options.body

        model = body.pop("model", None)
        stream = body.pop("stream", False)
        endpoint = "invoke"
        if stream:
            endpoint = "invoke-with-response-stream"

        # When stream is supported we would need to update this section and the URL
        url = f"{options.url}/model/{model}/{endpoint}"
        headers = self._prepare_headers(url=url, body=body)

        return options.replace(
            body=body,
            url=url,
            headers=headers,
        )

    def _prepare_headers(self, url: str, body: Dict[str, Any]) -> dict:
        return self._aws_auth.get_auth_headers(
            service_name="bedrock",
            method="POST",
            url=url,
            data=json.dumps(body),
        )


class AI21BedrockClient(AI21HTTPClient, BaseBedrockClient):
    def __init__(
        self,
        base_url: Optional[str] = None,
        region: Optional[str] = None,
        headers: Optional[Dict[str, Any]] = None,
        timeout_sec: Optional[float] = None,
        num_retries: Optional[int] = None,
        session: Optional[boto3.Session] = None,
        http_client: Optional[httpx.Client] = None,
    ):
        self._region = region or AI21EnvConfig.aws_region
        if base_url is None:
            base_url = _BEDROCK_URL_FORMAT.format(region=self._region)

        AI21HTTPClient.__init__(
            self,
            base_url=base_url,
            timeout_sec=timeout_sec,
            num_retries=num_retries,
            headers=headers,
            client=http_client,
            requires_api_key=False,
        )

        BaseBedrockClient.__init__(self, session=session, region=self._region)

        self.chat = StudioChat(self)

    def _build_request(self, options: RequestOptions) -> httpx.Request:
        options = self._prepare_options(options)

        return super()._build_request(options)

    def _prepare_url(self, options: RequestOptions) -> str:
        return options.url

    def _get_streaming_decoder(self) -> _AWSEventStreamDecoder:
        return self._streaming_decoder


class AsyncAI21BedrockClient(AsyncAI21HTTPClient, BaseBedrockClient):
    def __init__(
        self,
        base_url: Optional[str] = None,
        region: Optional[str] = None,
        headers: Optional[Dict[str, Any]] = None,
        timeout_sec: Optional[float] = None,
        num_retries: Optional[int] = None,
        session: Optional[boto3.Session] = None,
        http_client: Optional[httpx.AsyncClient] = None,
    ):
        self._region = region or AI21EnvConfig.aws_region
        if base_url is None:
            base_url = _BEDROCK_URL_FORMAT.format(region=self._region)

        AsyncAI21HTTPClient.__init__(
            self,
            base_url=base_url,
            timeout_sec=timeout_sec,
            num_retries=num_retries,
            headers=headers,
            client=http_client,
            requires_api_key=False,
        )

        BaseBedrockClient.__init__(self, session=session, region=self._region)

        self.chat = AsyncStudioChat(self)
        self.chat.create = self.chat.completions.create

    def _build_request(self, options: RequestOptions) -> httpx.Request:
        options = self._prepare_options(options)

        return super()._build_request(options)

    def _prepare_url(self, options: RequestOptions) -> str:
        return options.url

    def _get_streaming_decoder(self) -> _AWSEventStreamDecoder:
        return self._streaming_decoder



from ai21.models.chat.role_type import RoleType
from ai21.models.chat_message import ChatMessage
from ai21.models.document_type import DocumentType
from ai21.models.penalty import Penalty
from ai21.models.responses.chat_response import ChatOutput, ChatResponse, FinishReason
from ai21.models.responses.conversational_rag_response import (
    ConversationalRagResponse,
    ConversationalRagSource,
)
from ai21.models.responses.file_response import FileResponse


__all__ = [
    "ChatMessage",
    "RoleType",
    "Penalty",
    "DocumentType",
    "ChatResponse",
    "ChatOutput",
    "FinishReason",
    "FileResponse",
    "ConversationalRagResponse",
    "ConversationalRagSource",
]



from typing_extensions import TypedDict, Required

from ai21.models.chat.tool_parameters import ToolParameters


class FunctionToolDefinition(TypedDict, total=False):
    name: Required[str]
    description: str
    parameters: ToolParameters



from typing import Literal

from ai21.models.ai21_base_model import AI21BaseModel
from ai21.models.chat.tool_function import ToolFunction


class ToolCall(AI21BaseModel):
    id: str
    function: ToolFunction
    type: Literal["function"] = "function"
